#ifndef HASHGRID_H
#define HASHGRID_H

#include "hashtable.h"
#include "vec3f.h"

// the key for the hashtable: triples of integer grid coordinates
struct IntTriple
{
   int i, j, k;

   explicit IntTriple(int i_=0, int j_=0, int k_=0)
      : i(i_), j(j_), k(k_)
   {}

   bool operator==(const IntTriple &other) const { return i==other.i && j==other.j && k==other.k; }
};

unsigned int hash(const IntTriple &t);

// forward declaration of iterator
struct HashGrid3D_find_iterator;

// the hashtable itself, with extra stuff for searching neighbourhoods of a point
struct HashGrid3D
{
   float overdx, overdy, overdz; // reciprocals of size of each grid cell
   HashTable<IntTriple, int> grid; // hash table stores integer indices into array of particles
   const std::vector<Vec3f> &x; // reference to array containing positions of particles

   typedef HashGrid3D_find_iterator finder;

   explicit HashGrid3D(const std::vector<Vec3f> &x_, float dx=0.1)
      : overdx(1/dx), overdy(1/dx), overdz(1/dx), grid(x_.size()), x(x_)
   {
      // we won't actually add points to table until user calls update()
   }

   void update(void) // update hash table according to new positions in x
   {
      grid.clear();
      grid.reserve(x.size()); // to avoid unnecessary resizing during adds
      for(unsigned int p=0; p<x.size(); ++p)
         add_particle(p);
   }

   void add_particle(int index)
   { grid.add(IntTriple((int)(x[index][0]*overdx), (int)(x[index][1]*overdy), (int)(x[index][2]*overdz)), index); }

   void delete_particle(int index) // note - doesn't change x, just removes it from grid
   { grid.specific_delete(IntTriple((int)(x[index][0]*overdx), (int)(x[index][1]*overdy), (int)(x[index][2]*overdz)), index); }

   finder find(const Vec3f &centre, float radius) const;
};

// finds a superset of the points within a given radius of a point - up to user to throw out unwanted ones
struct HashGrid3D_find_iterator
{
   const HashGrid3D &hg;
   int ilow, ihigh, jlow, jhigh, klow, khigh;
   IntTriple cell;
   HashTable<IntTriple, int>::list_iterator *lister;

   HashGrid3D_find_iterator(const HashGrid3D &hg_, const Vec3f &centre, float radius);

   ~HashGrid3D_find_iterator()
   { delete lister; lister=0; }

   const int &operator*() const
   { return (*lister)->data; }

   void operator++();

   bool still_going() const
   { return lister!=0; }
};

#endif
