#ifndef HASHTABLE_H
#define HASHTABLE_H

#include <functional>
#include <vector>

template<class Key, class Data>
struct HashEntry
{
   Key key;
   Data data;
   int next;
};

// default hash function object
struct DefaultHashFunction
{
   template<typename Key>
   unsigned int operator() (const Key &k) const { return hash(k); }
};

struct equal
{
   template<typename T>
   bool operator() (const T &a, const T &b) const { return a==b; }
};

// forward declarations for iterators
template<typename Key, typename Data, class HashFunction, class KeyEqual> class HashTableIterator;
template<typename Key, typename Data, class HashFunction, class KeyEqual> class HashTableListIterator;

template<typename Key, typename Data, class HashFunction=DefaultHashFunction, class KeyEqual=equal>
struct HashTable
{
   unsigned int table_rank;
   unsigned int table_bits;
   std::vector<int> table;
   unsigned int num_entries;
   std::vector<HashEntry<Key, Data> > pool;
   int free_list;
   const HashFunction hash_function;
   const KeyEqual key_equal;

   typedef HashTableIterator<Key, Data, HashFunction, KeyEqual> iterator;
   typedef HashTableListIterator<Key, Data, HashFunction, KeyEqual> list_iterator;

   explicit HashTable(unsigned int expected_entries=64)
      : hash_function(HashFunction()), key_equal(KeyEqual())
   { init(expected_entries); }

   explicit HashTable(const HashFunction &hf, unsigned int expected_entries=64)
      : hash_function(hf), key_equal(KeyEqual())
   { init(expected_entries); }

   void init(unsigned int expected_entries)
   {
      unsigned int i;
      num_entries=0;
      table_rank=4;
      while(1u<<table_rank < expected_entries)
         ++table_rank;
      table_bits=(1u<<table_rank)-1;
      table.resize(1u<<table_rank);
      for(i=0; i<table.size(); ++i)
         table[i]=-1; // empty list
      pool.resize(1u<<table_rank);
      free_list=0;
      for(unsigned int i=0; i<pool.size()-1; ++i)
         pool[i].next=i+1;
      pool[pool.size()-1].next=-1; // end of free list
   }

   void add(const Key &k, const Data &d)
   {
      if(free_list==-1)
         reserve(1u<<(table_rank+1));
      int i=free_list; // where we're going to put the new entry
      free_list=pool[i].next;
      unsigned int t=hash_function(k)&table_bits; // index into table
      pool[i].key=k;
      pool[i].data=d;
      pool[i].next=table[t]; // put the new entry at the start of table[t]'s list
      table[t]=i;
      ++num_entries;
   }

   void specific_delete(const Key &k, const Data &d) // delete first entry that matches both key and data
   {
      unsigned int t=hash_function(k)&table_bits;
      int i=table[t], *p_i=&table[t];
      while(i!=-1){
	 if(key_equal(k, pool[i].key) && d==pool[i].data){
            *p_i=pool[i].next; // make list skip over this entry
	    pool[i].next=free_list; // and put it on the front of the free list
	    free_list=i;
	    return; // and we're done
	 }
	 p_i=&pool[i].next;
	 i=*p_i;
      }
   }

   unsigned int size() const
   { return num_entries; }

   void clear()
   {
      unsigned int i=0;
      num_entries=0;
      for(i=0; i<table.size(); ++i)
         table[i]=-1; // empty list
      free_list=0;
      for(i=0; i<pool.size()-1; ++i)
         pool[i].next=i+1;
      pool[pool.size()-1].next=-1;
   }

   void reserve(unsigned int expected_entries)
   {
      if(expected_entries<=pool.size())
         return;
      while(1u<<table_rank < expected_entries)
         ++table_rank;
      table_bits=(1u<<table_rank)-1;
      // increase room for new entries
      unsigned int old_size=pool.size(), i;
      pool.resize(1u<<table_rank);
      for(i=old_size; i<pool.size()-1; ++i)
         pool[i].next=i+1;
      pool[i].next=free_list;
      free_list=old_size;
      // And finally need to redo table (rehash entries)
      old_size=table.size();
      table.resize(1u<<table_rank);
      unsigned int t;
      for(t=old_size; t<table.size(); ++t)
         table[t]=-1; // initially make new lists empty
      int j, *p_j;
      for(t=0; t<old_size; ++t){
         j=table[t]; 
         p_j=&table[t];
         while(j!=-1){
            unsigned int new_t=hash_function(pool[j].key)&table_bits;
            if(new_t!=t){ // j doesn't belong in this list anymore?
               // delete from this list
               *p_j=pool[j].next;
               // add to correct list
               pool[j].next=table[new_t];
               table[new_t]=j;
            }else
	       p_j=&(pool[j].next);
            j=*p_j;
         }
      }
   }

   bool has_entry(const Key &k) const
   {
      unsigned int t=hash_function(k)&table_bits;
      int i=table[t];
      while(i!=-1){
         if(key_equal(k, pool[i].key))
            return true;
         i=pool[i].next;
      }
      return false;
   }

   bool get_entry(const Key &k, Data &data_return) const
   {
      unsigned int t=hash_function(k)&table_bits;
      int i=table[t];
      while(i!=-1){
         if(key_equal(k, pool[i].key)){
            data_return=pool[i].data;
            return true;
         }
         i=pool[i].next;
      }
      return false;
   }

   iterator begin(void) const // loop over table, looking for nonempty lists
   {
      if(num_entries==0)
         return end();
      unsigned int t=0;
      while(table[t]==-1) ++t;
      return iterator(*this, t, table[t]);
   }

   iterator end(void) const
   { return iterator(*this, table.size(), -1); }

   list_iterator begin_list(const Key &k) const
   { return list_iterator(*this, k); }
};

template<typename Key, typename Data, class HashFunction, class KeyEqual>
struct HashTableIterator
{
   const HashTable<Key, Data, HashFunction, KeyEqual> &hashtable;
   unsigned int t; // which table list we're working through
   int i; // which index into pool we're on

   HashTableIterator(const HashTable<Key, Data, HashFunction, KeyEqual> &ht, int t_, int i_)
      : hashtable(ht), t(t_), i(i_)
   {}

   const HashEntry<Key, Data> &operator*() const
   { return hashtable.pool[i]; }

   const HashEntry<Key, Data> *operator->() const
   { return &hashtable.pool[i]; }

   void operator++()
   {
      i=hashtable.pool[i].next;
      if(i==-1){
         for(;;){
            ++t;
            if(t==hashtable.table.size())
               break;
            else if(hashtable.table[t]!=-1){
               i=hashtable.table[t];
               break;
            }
         }
      }
   }

   bool operator!=(const HashTableIterator<Key, Data, HashFunction, KeyEqual> &other)
   { return &hashtable!=&other.hashtable || t!=other.t || i!=other.i; }
};

template<typename Key, typename Data, class HashFunction, class KeyEqual>
struct HashTableListIterator
{
   const HashTable<Key, Data, HashFunction, KeyEqual> &hashtable;
   const Key &key;
   int i; // which index into pool we're on

   HashTableListIterator(const HashTable<Key, Data, HashFunction, KeyEqual> &ht, const Key &k)
      : hashtable(ht), key(k)
   {
      i=hashtable.table[hashtable.hash_function(k)&hashtable.table_bits];
      while(i!=-1 && !hashtable.key_equal(hashtable.pool[i].key, key))
         i=hashtable.pool[i].next;
   }

   const HashEntry<Key, Data> &operator*() const
   { return hashtable.pool[i]; }

   const HashEntry<Key, Data> *operator->() const
   { return &hashtable.pool[i]; }

   void operator++()
   {
      do
         i=hashtable.pool[i].next;
      while(i!=-1 && !hashtable.key_equal(hashtable.pool[i].key, key));
   }

   bool still_going() const
   { return i!=-1; }
};

#endif
